# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""I{Base} classes representing XSD schema objects."""

from suds import *
from suds.xsd import *
from suds.sax.element import Element
from suds.sax import Namespace

from logging import getLogger
log = getLogger(__name__)


class SchemaObject(UnicodeMixin):
    """
    A schema object is an extension to object with schema awareness.

    @ivar root: The XML root element.
    @type root: L{Element}
    @ivar schema: The schema containing this object.
    @type schema: L{schema.Schema}
    @ivar form_qualified: A flag indicating that @elementFormDefault has a
        value of I{qualified}.
    @type form_qualified: boolean
    @ivar nillable: A flag indicating that @nillable has a value of I{true}.
    @type nillable: boolean
    @ivar default: The default value.
    @type default: object
    @ivar rawchildren: A list raw of all children.
    @type rawchildren: [L{SchemaObject},...]

    """

    @classmethod
    def prepend(cls, d, s, filter=Filter()):
        """
        Prepend B{s}ource XSD schema objects to the B{d}estination list.

        B{filter} is used to decide which objects to prepend and which to skip.

        @param d: The destination list.
        @type d: list
        @param s: The source list.
        @type s: list
        @param filter: A filter allowing items to be prepended.
        @type filter: L{Filter}

        """
        i = 0
        for x in s:
            if x in filter:
                d.insert(i, x)
                i += 1

    @classmethod
    def append(cls, d, s, filter=Filter()):
        """
        Append B{s}ource XSD schema objects to the B{d}estination list.

        B{filter} is used to decide which objects to append and which to skip.

        @param d: The destination list.
        @type d: list
        @param s: The source list.
        @type s: list
        @param filter: A filter that allows items to be appended.
        @type filter: L{Filter}

        """
        for item in s:
            if item in filter:
                d.append(item)

    def __init__(self, schema, root):
        """
        @param schema: The containing schema.
        @type schema: L{schema.Schema}
        @param root: The XML root node.
        @type root: L{Element}

        """
        self.schema = schema
        self.root = root
        self.id = objid(self)
        self.name = root.get("name")
        self.qname = (self.name, schema.tns[1])
        self.min = root.get("minOccurs")
        self.max = root.get("maxOccurs")
        self.type = root.get("type")
        self.ref = root.get("ref")
        self.form_qualified = schema.form_qualified
        self.nillable = False
        self.default = root.get("default")
        self.rawchildren = []

    def attributes(self, filter=Filter()):
        """
        Get only the attribute content.

        @param filter: A filter to constrain the result.
        @type filter: L{Filter}
        @return: A list of (attr, ancestry) tuples.
        @rtype: [(L{SchemaObject}, [L{SchemaObject},..]),..]

        """
        result = []
        for child, ancestry in self:
            if child.isattr() and child in filter:
                result.append((child, ancestry))
        return result

    def children(self, filter=Filter()):
        """
        Get only the I{direct} or non-attribute content.

        @param filter: A filter to constrain the result.
        @type filter: L{Filter}
        @return: A list tuples: (child, ancestry)
        @rtype: [(L{SchemaObject}, [L{SchemaObject},..]),..]

        """
        result = []
        for child, ancestry in self:
            if not child.isattr() and child in filter:
                result.append((child, ancestry))
        return result

    def get_attribute(self, name):
        """
        Get (find) an attribute by name.

        @param name: A attribute name.
        @type name: str
        @return: A tuple: the requested (attribute, ancestry).
        @rtype: (L{SchemaObject}, [L{SchemaObject},..])

        """
        for child, ancestry in self.attributes():
            if child.name == name:
                return child, ancestry
        return None, []

    def get_child(self, name):
        """
        Get (find) a I{non-attribute} child by name.

        @param name: A child name.
        @type name: str
        @return: A tuple: the requested (child, ancestry).
        @rtype: (L{SchemaObject}, [L{SchemaObject},..])

        """
        for child, ancestry in self.children():
            if child.any() or child.name == name:
                return child, ancestry
        return None, []

    def namespace(self, prefix=None):
        """
        Get this property's namespace.

        @param prefix: The default prefix.
        @type prefix: str
        @return: The schema's target namespace.
        @rtype: (I{prefix}, I{URI})

        """
        ns = self.schema.tns
        if ns[0] is None:
            ns = (prefix, ns[1])
        return ns

    def default_namespace(self):
        return self.root.defaultNamespace()

    def multi_occurrence(self):
        """
        Get whether the node has multiple occurrences, i.e. is a I{collection}.

        @return: True if it has, False if it has at most 1 occurrence.
        @rtype: boolean

        """
        max = self.max
        if max is None:
            return False
        if max.isdigit():
            return int(max) > 1
        return max == "unbounded"

    def optional(self):
        """
        Get whether this type is optional.

        @return: True if optional, else False.
        @rtype: boolean

        """
        return self.min == "0"

    def required(self):
        """
        Get whether this type is required.

        @return: True if required, else False.
        @rtype: boolean

        """
        return not self.optional()

    def resolve(self, nobuiltin=False):
        """
        Resolve the node's type reference and return the referenced type node.

        Only XSD schema objects that actually support 'having a type' custom
        implement this interface while others simply resolve as themselves.

        @param nobuiltin: Flag indicating whether resolving to an external XSD
            built-in type should not be allowed.
        @return: The resolved (true) type.
        @rtype: L{SchemaObject}
        """
        return self

    def sequence(self):
        """
        Get whether this is an <xsd:sequence/>.

        @return: True if <xsd:sequence/>, else False.
        @rtype: boolean

        """
        return False

    def xslist(self):
        """
        Get whether this is an <xsd:list/>.

        @return: True if <xsd:list/>, else False.
        @rtype: boolean

        """
        return False

    def all(self):
        """
        Get whether this is an <xsd:all/>.

        @return: True if <xsd:all/>, else False.
        @rtype: boolean

        """
        return False

    def choice(self):
        """
        Get whether this is an <xsd:choice/>.

        @return: True if <xsd:choice/>, else False.
        @rtype: boolean

        """
        return False

    def any(self):
        """
        Get whether this is an <xsd:any/>.

        @return: True if <xsd:any/>, else False.
        @rtype: boolean

        """
        return False

    def builtin(self):
        """
        Get whether this is a built-in schema-instance XSD type.

        @return: True if a built-in type, else False.
        @rtype: boolean

        """
        return False

    def enum(self):
        """
        Get whether this is a simple-type containing an enumeration.

        @return: True if enumeration, else False.
        @rtype: boolean

        """
        return False

    def isattr(self):
        """
        Get whether the object is a schema I{attribute} definition.

        @return: True if an attribute, else False.
        @rtype: boolean

        """
        return False

    def extension(self):
        """
        Get whether the object is an extension of another type.

        @return: True if an extension, else False.
        @rtype: boolean

        """
        return False

    def restriction(self):
        """
        Get whether the object is an restriction of another type.

        @return: True if a restriction, else False.
        @rtype: boolean

        """
        return False

    def mixed(self):
        """Get whether the object has I{mixed} content."""
        return False

    def find(self, qref, classes=[], ignore=None):
        """
        Find a referenced type in self or children. Return None if not found.

        Qualified references for all schema objects checked in this search will
        be added to the set of ignored qualified references to avoid the find
        operation going into an infinite loop in case of recursively defined
        structures.

        @param qref: A qualified reference.
        @type qref: qref
        @param classes: A collection of classes used to qualify the match.
        @type classes: Collection(I{class},...), e.g. [I(class),...]
        @param ignore: A set of qualified references to ignore in this search.
        @type ignore: {qref,...}
        @return: The referenced type.
        @rtype: L{SchemaObject}
        @see: L{qualify()}

        """
        if not len(classes):
            classes = (self.__class__,)
        if ignore is None:
            ignore = set()
        if self.qname in ignore:
            return
        ignore.add(self.qname)
        if self.qname == qref and self.__class__ in classes:
            return self
        for c in self.rawchildren:
            p = c.find(qref, classes, ignore=ignore)
            if p is not None:
                return p

    def translate(self, value, topython=True):
        """
        Translate between an XSD type values and Python objects.

        When converting a Python object to an XSD type value the operation may
        return any Python object whose string representation matches the
        desired XSD type value.

        @param value: A value to translate.
        @type value: str if topython is True; any Python object otherwise
        @param topython: Flag indicating the translation direction.
        @type topython: bool
        @return: The converted I{language} type.

        """
        return value

    def childtags(self):
        """
        Get a list of valid child tag names.

        @return: A list of child tag names.
        @rtype: [str,...]

        """
        return ()

    def dependencies(self):
        """
        Get a list of dependencies for dereferencing.

        @return: A merge dependency index and a list of dependencies.
        @rtype: (int, [L{SchemaObject},...])

        """
        return None, []

    def autoqualified(self):
        """
        The list of I{auto} qualified attribute values.

        Qualification means to convert values into I{qref}.

        @return: A list of attribute names.
        @rtype: list

        """
        return ["type", "ref"]

    def qualify(self):
        """
        Convert reference attribute values into a I{qref}.

        Constructed I{qref} uses the default document namespace. Since many
        WSDL schemas are written improperly: when the document does not define
        its default namespace, the schema target namespace is used to qualify
        references.

        """
        defns = self.root.defaultNamespace()
        if Namespace.none(defns):
            defns = self.schema.tns
        for a in self.autoqualified():
            ref = getattr(self, a)
            if ref is None:
                continue
            if isqref(ref):
                continue
            qref = qualify(ref, self.root, defns)
            log.debug("%s, convert %s='%s' to %s", self.id, a, ref, qref)
            setattr(self, a, qref)

    def merge(self, other):
        """Merge another object as needed."""
        other.qualify()
        for n in ("default", "max", "min", "name", "nillable", "qname",
                "type"):
            if getattr(self, n) is not None:
                continue
            v = getattr(other, n)
            if v is None:
                continue
            setattr(self, n, v)

    def content(self, collection=None, filter=Filter(), history=None):
        """
        Get a I{flattened} list of this node's contents.

        @param collection: A list to fill.
        @type collection: list
        @param filter: A filter used to constrain the result.
        @type filter: L{Filter}
        @param history: The history list used to prevent cyclic dependency.
        @type history: list
        @return: The filled list.
        @rtype: list

        """
        if collection is None:
            collection = []
        if history is None:
            history = []
        if self in history:
            return collection
        history.append(self)
        if self in filter:
            collection.append(self)
        for c in self.rawchildren:
            c.content(collection, filter, history)
        history.pop()
        return collection

    def str(self, indent=0, history=None):
        """
        Get a string representation of this object.

        @param indent: The indent.
        @type indent: int
        @return: A string.
        @rtype: str

        """
        if history is None:
            history = []
        if self in history:
            return "%s ..." % Repr(self)
        history.append(self)
        tab = "%*s" % (indent * 3, "")
        result = ["%s<%s" % (tab, self.id)]
        for n in self.description():
            if not hasattr(self, n):
                continue
            v = getattr(self, n)
            if v is None:
                continue
            result.append(' %s="%s"' % (n, v))
        if len(self):
            result.append(">")
            for c in self.rawchildren:
                result.append("\n")
                result.append(c.str(indent+1, history[:]))
                if c.isattr():
                    result.append("@")
            result.append("\n%s" % (tab,))
            result.append("</%s>" % (self.__class__.__name__,))
        else:
            result.append(" />")
        return "".join(result)

    def description(self):
        """
        Get the names used for repr() and str() description.

        @return:  A dictionary of relevant attributes.
        @rtype: [str,...]

        """
        return ()

    def __unicode__(self):
        return str(self.str())

    def __repr__(self):
        s = []
        s.append("<%s" % (self.id,))
        for n in self.description():
            if not hasattr(self, n):
                continue
            v = getattr(self, n)
            if v is None:
                continue
            s.append(' %s="%s"' % (n, v))
        s.append(" />")
        return "".join(s)

    def __len__(self):
        n = 0
        for x in self:
            n += 1
        return n

    def __iter__(self):
        return Iter(self)

    def __getitem__(self, index):
        """
        Returns a contained schema object referenced by its 0-based index.

        Returns None if such an object does not exist.

        """
        i = 0
        for c in self:
            if i == index:
                return c
            i += 1


class Iter:
    """
    The content iterator - used to iterate the L{Content} children.

    The iterator provides a I{view} of the children that is free of container
    elements such as <xsd::all/>, <xsd:choice/> or <xsd:sequence/>.

    @ivar stack: A stack used to control nesting.
    @type stack: list

    """

    class Frame:
        """A content iterator frame."""

        def __init__(self, sx):
            """
            @param sx: A schema object.
            @type sx: L{SchemaObject}

            """
            self.sx = sx
            self.items = sx.rawchildren
            self.index = 0

        def __next__(self):
            """
            Get the I{next} item in the frame's collection.

            @return: The next item or None
            @rtype: L{SchemaObject}

            """
            if self.index < len(self.items):
                result = self.items[self.index]
                self.index += 1
                return result

    def __init__(self, sx):
        """
        @param sx: A schema object.
        @type sx: L{SchemaObject}

        """
        self.stack = []
        self.push(sx)

    def push(self, sx):
        """
        Create a frame and push the specified object.

        @param sx: A schema object to push.
        @type sx: L{SchemaObject}

        """
        self.stack.append(Iter.Frame(sx))

    def pop(self):
        """
        Pop the I{top} frame.

        @return: The popped frame.
        @rtype: L{Frame}
        @raise StopIteration: when stack is empty.

        """
        if self.stack:
            return self.stack.pop()
        raise StopIteration()

    def top(self):
        """
        Get the I{top} frame.

        @return: The top frame.
        @rtype: L{Frame}
        @raise StopIteration: when stack is empty.

        """
        if self.stack:
            return self.stack[-1]
        raise StopIteration()

    def __next__(self):
        """
        Get the next item.

        @return: A tuple: the next (child, ancestry).
        @rtype: (L{SchemaObject}, [L{SchemaObject},..])
        @raise StopIteration: A the end.

        """
        frame = self.top()
        while True:
            result = next(frame)
            if result is None:
                self.pop()
                return next(self)
            if isinstance(result, Content):
                ancestry = [f.sx for f in self.stack]
                return result, ancestry
            self.push(result)
            return next(self)

    def __iter__(self):
        return self


class XBuiltin(SchemaObject):
    """Represents a built-in XSD schema <xsd:*/> node."""

    def __init__(self, schema, name):
        """
        @param schema: The containing schema.
        @type schema: L{schema.Schema}

        """
        root = Element(name)
        SchemaObject.__init__(self, schema, root)
        self.name = name
        self.nillable = True

    def namespace(self, prefix=None):
        return Namespace.xsdns

    def builtin(self):
        return True


class Content(SchemaObject):
    """XSD schema objects representing real XML document content."""
    pass


class NodeFinder:
    """
    Find nodes based on flexable criteria.

    I{matcher} may be any object implementing a match(n) method.

    @ivar matcher: An object used as criteria for match.
    @type matcher: I{any}.match(n)
    @ivar limit: Limit the number of matches.  0=unlimited.
    @type limit: int

    """
    def __init__(self, matcher, limit=0):
        """
        @param matcher: An object used as criteria for match.
        @type matcher: I{any}.match(n)
        @param limit: Limit the number of matches. 0=unlimited.
        @type limit: int

        """
        self.matcher = matcher
        self.limit = limit

    def find(self, node, list):
        """
        Traverse the tree looking for matches.

        @param node: A node to match on.
        @type node: L{SchemaObject}
        @param list: A list to fill.
        @type list: list

        """
        if self.matcher.match(node):
            list.append(node)
            self.limit -= 1
            if self.limit == 0:
                return
        for c in node.rawchildren:
            self.find(c, list)
        return self



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jurko Gospodnetić ( jurko.gospodnetic@pke.hr )

"""
"poor man's tox" development script used on Windows to run the full suds-jurko
test suite using multiple Python interpreter versions.

Intended to be used as a general 'all tests passed' check. To see more detailed
information on specific failures, run the failed test group manually,
configured for greater verbosity than done here.

"""

import os.path
import shutil
import sys

from suds_devel.configuration import BadConfiguration, Config, configparser
from suds_devel.environment import BadEnvironment
import suds_devel.utility as utility


class MyConfig(Config):

    def __init__(self, script, project_folder, ini_file):
        """
        Initialize new script configuration.

        External configuration parameters may be specified relative to the
        following folders:
          * script - relative to the current working folder
          * project_folder - relative to the script folder
          * ini_file - relative to the project folder

        """
        super(MyConfig, self).__init__(script, project_folder, ini_file)
        try:
            self._read_environment_configuration()
        except configparser.Error:
            raise BadConfiguration(sys.exc_info()[1].message)


def _prepare_configuration():
    # We know we are a regular stand-alone script file and not an imported
    # module (either frozen, imported from disk, zip-file, external database or
    # any other source). That means we can safely assume we have the __file__
    # attribute available.
    global config
    config = MyConfig(__file__, "..", "setup.cfg")


def _print_title(env, message_fmt):
    separator = "-" * 63
    print("")
    print(separator)
    print("--- " + message_fmt % (env.name(),))
    print(separator)


def _report_startup_information():
    print("Running in folder: '%s'" % (os.getcwd(),))


def _run_tests(env):
    if env.sys_version_info >= (3,):
        _print_title(env, "Building suds for Python %s")
        build_folder = os.path.join(config.project_folder, "build")
        if os.path.isdir(build_folder):
            shutil.rmtree(build_folder)

    # Install the project into the target Python environment in editable mode.
    # This will actually build Python 3 sources in case we are using a Python 3
    # environment.
    setup_cmd = ["setup.py", "-q", "develop"]
    _, _, return_code = env.execute(setup_cmd, cwd=config.project_folder)
    if return_code != 0:
        return False

    test_folder = os.path.join(config.project_folder, "tests")
    pytest_cmd = ["-m", "pytest", "-q", "-x", "--tb=short"]

    _print_title(env, "Testing suds with Python %s")
    _, _, return_code = env.execute(pytest_cmd, cwd=test_folder)
    if return_code != 0:
        return False

    _print_title(env, "Testing suds with Python %s - no assertions")
    pytest_cmd.insert(0, "-O")
    _, _, return_code = env.execute(pytest_cmd, cwd=test_folder)
    return return_code == 0


def _run_tests_in_all_environments():
    if not config.python_environments:
        raise BadConfiguration("No Python environments configured.")
    for env in config.python_environments:
        if not env.initial_scan_completed:
            _print_title(env, "Scanning environment Python %s")
            env.run_initial_scan()
        if not _run_tests(env):
            return False
    return True


def main():
    try:
        _report_startup_information()
        _prepare_configuration()
        success = _run_tests_in_all_environments()
    except (BadConfiguration, BadEnvironment):
        utility.report_error(sys.exc_info()[1])
        return -2
    print("")
    if not success:
        print("Test failed.")
        return -3
    print("All tests passed.")
    return 0


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jurko Gospodnetić ( jurko.gospodnetic@pke.hr )

"""
Suds SAX Element unit tests.

Implemented using the 'pytest' testing framework.

"""

if __name__ == "__main__":
    import testutils
    testutils.run_using_pytest(globals())

import suds
from suds.sax.element import Element
import suds.sax.parser

import pytest
import six

import re
import sys


class TestChildAtPath:

    def test_backslash_as_path_separator(self):
        name1 = "child"
        name2 = "grandchild"
        root = self.__create_single_branch("root", name1, name2)[0]
        result = root.childAtPath(name1 + "\\" + name2)
        assert result is None

    def test_backslash_in_name(self):
        root, a, _, _ = self.__create_single_branch("root", "a", "b", "c")
        b_c = Element("b\\c")
        a.append(b_c)
        result = root.childAtPath("a/b\\c")
        assert result is b_c

    def test_child_leaf(self):
        root, child = self.__create_single_branch("root", "child")
        result = root.childAtPath("child")
        assert result is child

    def test_child_not_leaf(self):
        root, child, _ = self.__create_single_branch("root", "child",
            "grandchild")
        result = root.childAtPath("child")
        assert result is child

    def test_grandchild_leaf(self):
        root, _, grandchild = self.__create_single_branch("root", "child",
            "grandchild")
        result = root.childAtPath("child/grandchild")
        assert result is grandchild

    def test_grandchild_not_leaf(self):
        root, _, grandchild, _ = self.__create_single_branch("root", "child",
            "grandchild", "great grandchild")
        result = root.childAtPath("child/grandchild")
        assert result is grandchild

    def test_misplaced(self):
        root = self.__create_single_branch("root", "a", "x", "b")[0]
        result = root.childAtPath("a/b")
        assert result is None

    def test_missing(self):
        root = Element("root")
        result = root.childAtPath("an invalid path")
        assert result is None

    def test_name_including_spaces(self):
        root, _, child, _ = self.__create_single_branch("root", "dumbo",
            "foo  -  bar", "baz")
        result = root.childAtPath("dumbo/foo  -  bar")
        assert result is child

    @pytest.mark.parametrize("n", (2, 3))
    def test_repeated_path_separators(self, n):
        root, child, grandchild = self.__create_single_branch("root", "child",
            "grandchild")
        sep = "/" * n
        path = "child" + sep + "grandchild"
        result = root.childAtPath(path)
        assert result is grandchild

    def test_same_named(self):
        root, _, child, _ = self.__create_single_branch("root", "a", "a", "a")
        result = root.childAtPath("a/a")
        assert result is child

    @staticmethod
    def __create_single_branch(*args):
        """
        Construct a single branch element tree with given element names.

        Returns a list of constructed Element nodes from root to leaf.

        """
        result = []
        parent = None
        for name in args:
            e = Element(name)
            result.append(e)
            if parent is not None:
                parent.append(e)
            parent = e
        return result


class TestStringRepresentation:

    # Must be consistent with how Element.str() formats this data.
    str_formatted_xml = """\
<xsd:element name="ZuZu">
   <xsd:simpleType>
      <xsd:restriction base="xsd:string">
         <xsd:enumeration value="alfa"/>
         <xsd:enumeration value="beta"/>
         <xsd:enumeration value="gamma"/>
      </xsd:restriction>
   </xsd:simpleType>
</xsd:element>"""

    @staticmethod
    def create_test_element(content=str_formatted_xml):
        input_data = suds.byte_str(content)
        xml = suds.sax.parser.Parser().parse(suds.BytesIO(input_data))
        element = xml.root()
        assert element.__class__ is Element
        return element

    @pytest.mark.skipif(sys.version_info >= (3,), reason="Python 2 specific")
    def test_convert_to_byte_str(self):
        element = self.create_test_element()
        expected = suds.byte_str(element.str())
        assert str(element) == expected

    def test_convert_to_unicode(self):
        element = self.create_test_element()
        expected = element.str()
        assert six.text_type(element) == expected

    def test_plain_method(self):
        element = self.create_test_element(self.str_formatted_xml)
        expected = re.sub(r"\s*[\r\n]\s*", "", self.str_formatted_xml)
        result = element.plain()
        assert result == expected

    def test_str_method(self):
        element = self.create_test_element(self.str_formatted_xml)
        result = element.str()
        assert result == self.str_formatted_xml


@pytest.mark.parametrize("name, expected_prefix, expected_name", (
    ("", None, ""),
    ("bazinga", None, "bazinga"),
    ("test element name", None, "test element name"),
    ("aaa:bbb", "aaa", "bbb"),
    ("aaa:", "aaa", ""),
    (":aaa", "", "aaa"),
    ("aaa::bbb", "aaa", ":bbb"),
    ("aaa:bbb:ccc", "aaa", "bbb:ccc")))
def test_init_name(name, expected_prefix, expected_name):
    e = Element(name)
    assert e.prefix == expected_prefix
    assert e.name == expected_name



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jurko Gospodnetić ( jurko.gospodnetic@pke.hr )

"""
Suds MX system unit tests.

Implemented using the 'pytest' testing framework.

"""

if __name__ == "__main__":
    import testutils
    testutils.run_using_pytest(globals())

from suds.mx.typer import Typer

import pytest


def _prefix(i):
    """Prefixes expected to be constructed by Typer.getprefix()."""
    return "ns%d" % (i,)


class TestTyper:

    class Test_genprefix:

        class MockNode:

            def __init__(self, namespace_mapping):
                self.mock_call_params = []
                self.__namespace_mapping = namespace_mapping

            def resolvePrefix(self, prefix, default):
                self.mock_call_params.append((prefix, default))
                return self.__namespace_mapping.get(prefix, default)

        def test_no_mapped_prefixes(self):
            node = self.__class__.MockNode({})
            test_namespace = "test namespace"
            result = Typer.genprefix(node, ("unused-prefix", test_namespace))
            assert result == (_prefix(1), test_namespace)
            assert node.mock_call_params == [(_prefix(1), None)]

        def test_several_already_mapped_prefixes(self):
            test_namespace = "test namespace"
            node = self.__class__.MockNode({
                _prefix(1): "another namespace",
                _prefix(2): "another namespace"})
            result = Typer.genprefix(node, ("unused-prefix", test_namespace))
            assert result == (_prefix(3), test_namespace)
            assert node.mock_call_params == [
                (_prefix(i), None) for i in [1, 2, 3]]

        def test_last_free_namespace(self):
            test_namespace = "test namespace"
            node = self.__class__.MockNode(dict(
                (_prefix(i), "another namespace")
                for i in range(1, 1023)))
            result = Typer.genprefix(node, ("unused-prefix", test_namespace))
            assert result == (_prefix(1023), test_namespace)
            expected_calls = [(_prefix(i), None) for i in range(1, 1024)]
            assert node.mock_call_params == expected_calls

        def test_no_free_namespace(self):
            test_namespace = "test namespace"
            node = self.__class__.MockNode(dict(
                (_prefix(i), "another namespace") for i in range(1, 1024)))
            e = pytest.raises(Exception,
                Typer.genprefix, node, ("unused-prefix", test_namespace)).value
            try:
                assert str(e) == "auto prefix, exhausted"
            finally:
                del e  # explicitly break circular reference chain in Python 3
            expected_calls = [(_prefix(i), None) for i in range(1, 1024)]
            assert node.mock_call_params == expected_calls

        def test_already_mapped_namespace_with_no_unused_prefix_before(self):
            test_prefix_index = 2
            test_namespace = "test namespace"
            node = self.__class__.MockNode({
                _prefix(1): "another namespace",
                _prefix(test_prefix_index): test_namespace,
                _prefix(3): "another namespace"})
            result = Typer.genprefix(node, ("unused-prefix", test_namespace))
            assert result == (_prefix(test_prefix_index), test_namespace)
            expected_calls = [(_prefix(i), None)
                for i in range(1, test_prefix_index + 1)]
            assert node.mock_call_params == expected_calls

        def test_already_mapped_namespace_with_unused_prefix_before(self):
            unused_prefix_index = 2
            test_namespace = "test namespace"
            node = self.__class__.MockNode({
                _prefix(1): "another namespace",
                _prefix(3): test_namespace,
                _prefix(4): "another namespace"})
            result = Typer.genprefix(node, ("unused-prefix", test_namespace))
            assert result == (_prefix(unused_prefix_index), test_namespace)
            expected_calls = [(_prefix(i), None)
                for i in range(1, unused_prefix_index + 1)]
            assert node.mock_call_params == expected_calls



# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""
Provides XML I{special character} encoder classes.

"""

import re


class Encoder:
    """
    An XML special character encoder/decoder.

    @cvar encodings: A mapping of special characters encoding.
    @type encodings: [(str, str),...]
    @cvar decodings: A mapping of special characters decoding.
    @type decodings: [(str, str),...]
    @cvar special: A list of special characters.
    @type special: [char,...]

    """

    encodings = (
        ("&(?!(amp|lt|gt|quot|apos);)", "&amp;"),
        ("<", "&lt;"),
        (">", "&gt;"),
        ('"', "&quot;"),
        ("'", "&apos;"))
    decodings = (
        ("&lt;", "<"),
        ("&gt;", ">"),
        ("&quot;", '"'),
        ("&apos;", "'"),
        ("&amp;", "&"))
    special = ("&", "<", ">", '"', "'")

    def encode(self, s):
        """
        Encode special characters found in string I{s}.

        @param s: A string to encode.
        @type s: str
        @return: The encoded string.
        @rtype: str

        """
        if isinstance(s, str) and self.__needs_encoding(s):
            for x in self.encodings:
                s = re.sub(x[0], x[1], s)
        return s

    def decode(self, s):
        """
        Decode special characters encodings found in string I{s}.

        @param s: A string to decode.
        @type s: str
        @return: The decoded string.
        @rtype: str

        """
        if isinstance(s, str) and "&" in s:
            for x in self.decodings:
                s = s.replace(x[0], x[1])
        return s

    def __needs_encoding(self, s):
        """
        Get whether string I{s} contains special characters.

        @param s: A string to check.
        @type s: str
        @return: True if needs encoding.
        @rtype: boolean

        """
        if isinstance(s, str):
            for c in self.special:
                if c in s:
                    return True



# This program is free software; you can redistribute it and/or modify
# it under the terms of the (LGPL) GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library Lesser General Public License for more details at
# ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""
Provides filtered attribute list classes.
"""

from suds import *
from suds.umx import *
from suds.sax import Namespace


class AttrList:
    """
    A filtered attribute list.
    Items are included during iteration if they are in either the (xs) or
    (xml) namespaces.
    @ivar raw: The I{raw} attribute list.
    @type raw: list
    """
    def __init__(self, attributes):
        """
        @param attributes: A list of attributes
        @type attributes: list
        """
        self.raw = attributes

    def real(self):
        """
        Get list of I{real} attributes which exclude xs and xml attributes.
        @return: A list of I{real} attributes.
        @rtype: I{generator}
        """
        for a in self.raw:
            if self.skip(a): continue
            yield a

    def rlen(self):
        """
        Get the number of I{real} attributes which exclude xs and xml attributes.
        @return: A count of I{real} attributes.
        @rtype: L{int}
        """
        n = 0
        for a in self.real():
            n += 1
        return n

    def lang(self):
        """
        Get list of I{filtered} attributes which exclude xs.
        @return: A list of I{filtered} attributes.
        @rtype: I{generator}
        """
        for a in self.raw:
            if a.qname() == 'xml:lang':
                return a.value
            return None

    def skip(self, attr):
        """
        Get whether to skip (filter-out) the specified attribute.
        @param attr: An attribute.
        @type attr: I{Attribute}
        @return: True if should be skipped.
        @rtype: bool
        """
        ns = attr.namespace()
        skip = (
            Namespace.xmlns[1],
            'http://schemas.xmlsoap.org/soap/encoding/',
            'http://schemas.xmlsoap.org/soap/envelope/',
            'http://www.w3.org/2003/05/soap-envelope',
        )
        return ( Namespace.xs(ns) or ns[1] in skip )



# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""
The I{schema} module provides an intelligent representation of an XSD schema.
The I{raw} model is the XML tree and the I{model} is a denormalized,
objectified and intelligent view of the schema. Most of the I{value-add}
provided by the model is centered around transparent referenced type resolution
and targeted denormalization.

"""

from suds import *
from suds.xsd import *
from suds.xsd.depsort import dependency_sort
from suds.xsd.sxbuiltin import *
from suds.xsd.sxbase import SchemaObject
from suds.xsd.sxbasic import Factory as BasicFactory
from suds.xsd.sxbuiltin import Factory as BuiltinFactory
from suds.sax import splitPrefix, Namespace
from suds.sax.element import Element

from logging import getLogger
log = getLogger(__name__)


class SchemaCollection(UnicodeMixin):
    """
    A collection of schema objects.

    This class is needed because a WSDL may contain more then one <schema/>
    node.

    @ivar wsdl: A WSDL object.
    @type wsdl: L{suds.wsdl.Definitions}
    @ivar children: A list contained schemas.
    @type children: [L{Schema},...]
    @ivar namespaces: A dictionary of contained schemas by namespace.
    @type namespaces: {str: L{Schema}}

    """

    def __init__(self, wsdl):
        """
        @param wsdl: A WSDL object.
        @type wsdl: L{suds.wsdl.Definitions}

        """
        self.wsdl = wsdl
        self.children = []
        self.namespaces = {}

    def add(self, schema):
        """
        Add a schema node to the collection. Schema(s) within the same target
        namespace are consolidated.

        @param schema: A schema object.
        @type schema: (L{Schema})

        """
        key = schema.tns[1]
        existing = self.namespaces.get(key)
        if existing is None:
            self.children.append(schema)
            self.namespaces[key] = schema
        else:
            existing.root.children += schema.root.children
            existing.root.nsprefixes.update(schema.root.nsprefixes)

    def load(self, options, loaded_schemata):
        """
        Load schema objects for the root nodes.
            - de-reference schemas
            - merge schemas

        @param options: An options dictionary.
        @type options: L{options.Options}
        @param loaded_schemata: Already loaded schemata cache (URL --> Schema).
        @type loaded_schemata: dict
        @return: The merged schema.
        @rtype: L{Schema}

        """
        if options.autoblend:
            self.autoblend()
        for child in self.children:
            child.build()
        for child in self.children:
            child.open_imports(options, loaded_schemata)
        for child in self.children:
            child.dereference()
        log.debug("loaded:\n%s", self)
        merged = self.merge()
        log.debug("MERGED:\n%s", merged)
        return merged

    def autoblend(self):
        """
        Ensure that all schemas within the collection import each other which
        has a blending effect.

        @return: self
        @rtype: L{SchemaCollection}

        """
        namespaces = list(self.namespaces.keys())
        for s in self.children:
            for ns in namespaces:
                tns = s.root.get("targetNamespace")
                if tns == ns:
                    continue
                for imp in s.root.getChildren("import"):
                    if imp.get("namespace") == ns:
                        continue
                imp = Element("import", ns=Namespace.xsdns)
                imp.set("namespace", ns)
                s.root.append(imp)
        return self

    def locate(self, ns):
        """
        Find a schema by namespace. Only the URI portion of the namespace is
        compared to each schema's I{targetNamespace}.

        @param ns: A namespace.
        @type ns: (prefix, URI)
        @return: The schema matching the namespace, else None.
        @rtype: L{Schema}

        """
        return self.namespaces.get(ns[1])

    def merge(self):
        """
        Merge contained schemas into one.

        @return: The merged schema.
        @rtype: L{Schema}

        """
        if self.children:
            schema = self.children[0]
            for s in self.children[1:]:
                schema.merge(s)
            return schema

    def __len__(self):
        return len(self.children)

    def __unicode__(self):
        result = ["\nschema collection"]
        for s in self.children:
            result.append(s.str(1))
        return "\n".join(result)


class Schema(UnicodeMixin):
    """
    The schema is an objectification of a <schema/> (XSD) definition. It
    provides inspection, lookup and type resolution.

    @ivar root: The root node.
    @type root: L{sax.element.Element}
    @ivar baseurl: The I{base} URL for this schema.
    @type baseurl: str
    @ivar container: A schema collection containing this schema.
    @type container: L{SchemaCollection}
    @ivar children: A list of direct top level children.
    @type children: [L{SchemaObject},...]
    @ivar all: A list of all (includes imported) top level children.
    @type all: [L{SchemaObject},...]
    @ivar types: A schema types cache.
    @type types: {name:L{SchemaObject}}
    @ivar imports: A list of import objects.
    @type imports: [L{SchemaObject},...]
    @ivar elements: A list of <element/> objects.
    @type elements: [L{SchemaObject},...]
    @ivar attributes: A list of <attribute/> objects.
    @type attributes: [L{SchemaObject},...]
    @ivar groups: A list of group objects.
    @type groups: [L{SchemaObject},...]
    @ivar agrps: A list of attribute group objects.
    @type agrps: [L{SchemaObject},...]
    @ivar form_qualified: The flag indicating: (@elementFormDefault).
    @type form_qualified: bool

    """

    Tag = "schema"

    def __init__(self, root, baseurl, options, loaded_schemata=None,
            container=None):
        """
        @param root: The XML root.
        @type root: L{sax.element.Element}
        @param baseurl: The base URL used for importing.
        @type baseurl: basestring
        @param options: An options dictionary.
        @type options: L{options.Options}
        @param loaded_schemata: An optional already loaded schemata cache (URL
            --> Schema).
        @type loaded_schemata: dict
        @param container: An optional container.
        @type container: L{SchemaCollection}

        """
        self.root = root
        self.id = objid(self)
        self.tns = self.mktns()
        self.baseurl = baseurl
        self.container = container
        self.children = []
        self.all = []
        self.types = {}
        self.imports = []
        self.elements = {}
        self.attributes = {}
        self.groups = {}
        self.agrps = {}
        if options.doctor is not None:
            options.doctor.examine(root)
        form = self.root.get("elementFormDefault")
        self.form_qualified = form == "qualified"

        # If we have a container, that container is going to take care of
        # finishing our build for us in parallel with building all the other
        # schemata in that container. That allows the different schema within
        # the same container to freely reference each other.
        #TODO: check whether this container content build parallelization is
        # really necessary or if we can simply build our top-level WSDL
        # contained schemata one by one as they are loaded
        if container is None:
            if loaded_schemata is None:
                loaded_schemata = {}
            loaded_schemata[baseurl] = self
            #TODO: It seems like this build() step can be done for each schema
            # on its own instead of letting the container do it. Building our
            # XSD schema objects should not require any external schema
            # information and even references between XSD schema objects within
            # the same schema can not be established until all the XSD schema
            # objects have been built. The only reason I can see right now why
            # this step has been placed under container control is so our
            # container (a SchemaCollection instance) can add some additional
            # XML elements to our schema before our XSD schema object entities
            # get built, but there is bound to be a cleaner way to do this,
            # similar to how we support such XML modifications in suds plugins.
            self.build()
            self.open_imports(options, loaded_schemata)
            log.debug("built:\n%s", self)
            self.dereference()
            log.debug("dereferenced:\n%s", self)

    def mktns(self):
        """
        Make the schema's target namespace.

        @return: namespace representation of the schema's targetNamespace
            value.
        @rtype: (prefix, URI)

        """
        tns = self.root.get("targetNamespace")
        tns_prefix = None
        if tns is not None:
            tns_prefix = self.root.findPrefix(tns)
        return tns_prefix, tns

    def build(self):
        """
        Build the schema (object graph) using the root node using the factory.
            - Build the graph.
            - Collate the children.

        """
        self.children = BasicFactory.build(self.root, self)
        collated = BasicFactory.collate(self.children)
        self.children = collated[0]
        self.attributes = collated[2]
        self.imports = collated[1]
        self.elements = collated[3]
        self.types = collated[4]
        self.groups = collated[5]
        self.agrps = collated[6]

    def merge(self, schema):
        """
        Merge the schema contents.

        Only objects not already contained in this schema's collections are
        merged. This provides support for bidirectional imports producing
        cyclic includes.

        @returns: self
        @rtype: L{Schema}

        """
        for item in list(schema.attributes.items()):
            if item[0] in self.attributes:
                continue
            self.all.append(item[1])
            self.attributes[item[0]] = item[1]
        for item in list(schema.elements.items()):
            if item[0] in self.elements:
                continue
            self.all.append(item[1])
            self.elements[item[0]] = item[1]
        for item in list(schema.types.items()):
            if item[0] in self.types:
                continue
            self.all.append(item[1])
            self.types[item[0]] = item[1]
        for item in list(schema.groups.items()):
            if item[0] in self.groups:
                continue
            self.all.append(item[1])
            self.groups[item[0]] = item[1]
        for item in list(schema.agrps.items()):
            if item[0] in self.agrps:
                continue
            self.all.append(item[1])
            self.agrps[item[0]] = item[1]
        schema.merged = True
        return self

    def open_imports(self, options, loaded_schemata):
        """
        Instruct all contained L{sxbasic.Import} children to import all of
        their referenced schemas. The imported schema contents are I{merged}
        in.

        @param options: An options dictionary.
        @type options: L{options.Options}
        @param loaded_schemata: Already loaded schemata cache (URL --> Schema).
        @type loaded_schemata: dict

        """
        for imp in self.imports:
            imported = imp.open(options, loaded_schemata)
            if imported is None:
                continue
            imported.open_imports(options, loaded_schemata)
            log.debug("imported:\n%s", imported)
            self.merge(imported)

    def dereference(self):
        """Instruct all children to perform dereferencing."""
        all = []
        indexes = {}
        for child in self.children:
            child.content(all)
        dependencies = {}
        for x in all:
            x.qualify()
            midx, deps = x.dependencies()
            dependencies[x] = deps
            indexes[x] = midx
        for x, deps in dependency_sort(dependencies):
            midx = indexes.get(x)
            if midx is None:
                continue
            d = deps[midx]
            log.debug("(%s) merging %s <== %s", self.tns[1], Repr(x), Repr(d))
            x.merge(d)

    def locate(self, ns):
        """
        Find a schema by namespace. Only the URI portion of the namespace is
        compared to each schema's I{targetNamespace}. The request is passed on
        to the container.

        @param ns: A namespace.
        @type ns: (prefix, URI)
        @return: The schema matching the namespace, else None.
        @rtype: L{Schema}

        """
        if self.container is not None:
            return self.container.locate(ns)

    def custom(self, ref, context=None):
        """
        Get whether the specified reference is B{not} an (xs) builtin.

        @param ref: A str or qref.
        @type ref: (str|qref)
        @return: True if B{not} a builtin, else False.
        @rtype: bool

        """
        return ref is None or not self.builtin(ref, context)

    def builtin(self, ref, context=None):
        """
        Get whether the specified reference is an (xs) builtin.

        @param ref: A str or qref.
        @type ref: (str|qref)
        @return: True if builtin, else False.
        @rtype: bool

        """
        w3 = "http://www.w3.org"
        try:
            if isqref(ref):
                ns = ref[1]
                return ref[0] in Factory.tags and ns.startswith(w3)
            if context is None:
                context = self.root
            prefix = splitPrefix(ref)[0]
            prefixes = context.findPrefixes(w3, "startswith")
            return prefix in prefixes and ref[0] in Factory.tags
        except Exception:
            return False

    def instance(self, root, baseurl, loaded_schemata, options):
        """
        Create and return an new schema object using the specified I{root} and
        I{URL}.

        @param root: A schema root node.
        @type root: L{sax.element.Element}
        @param baseurl: A base URL.
        @type baseurl: str
        @param loaded_schemata: Already loaded schemata cache (URL --> Schema).
        @type loaded_schemata: dict
        @param options: An options dictionary.
        @type options: L{options.Options}
        @return: The newly created schema object.
        @rtype: L{Schema}
        @note: This is only used by Import children.

        """
        return Schema(root, baseurl, options, loaded_schemata)

    def str(self, indent=0):
        tab = "%*s" % (indent * 3, "")
        result = []
        result.append("%s%s" % (tab, self.id))
        result.append("%s(raw)" % (tab,))
        result.append(self.root.str(indent + 1))
        result.append("%s(model)" % (tab,))
        for c in self.children:
            result.append(c.str(indent + 1))
        result.append("")
        return "\n".join(result)

    def __repr__(self):
        return '<%s tns="%s"/>' % (self.id, self.tns[1])

    def __unicode__(self):
        return self.str()



# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""
Classes providing a (D)ocument (O)bject (M)odel representation of an XML
document.

The goal is to provide an easy, intuitive interface for managing XML documents.
Although the term DOM is used above, this model is B{far} better.

XML namespaces in suds are represented using a (2) element tuple containing the
prefix and the URI, e.g. I{('tns', 'http://myns')}.

"""

import suds
from suds import *
from suds.sax import *
from suds.sax.attribute import Attribute
from suds.sax.document import Document
from suds.sax.element import Element
from suds.sax.text import Text

import sys
from xml.sax import make_parser, InputSource, ContentHandler
from xml.sax.handler import feature_external_ges


class Handler(ContentHandler):
    """SAX handler."""

    def __init__(self):
        self.nodes = [Document()]

    def startElement(self, name, attrs):
        top = self.top()
        node = Element(str(name))
        for a in attrs.getNames():
            n = str(a)
            v = str(attrs.getValue(a))
            attribute = Attribute(n, v)
            if self.mapPrefix(node, attribute):
                continue
            node.append(attribute)
        node.charbuffer = []
        top.append(node)
        self.push(node)

    def mapPrefix(self, node, attribute):
        if attribute.name == "xmlns":
            if len(attribute.value):
                node.expns = str(attribute.value)
            return True
        if attribute.prefix == "xmlns":
            prefix = attribute.name
            node.nsprefixes[prefix] = str(attribute.value)
            return True
        return False

    def endElement(self, name):
        name = str(name)
        current = self.pop()
        if name != current.qname():
            raise Exception("malformed document")
        if current.charbuffer:
            current.text = Text("".join(current.charbuffer))
        del current.charbuffer
        if current:
            current.trim()

    def characters(self, content):
        text = str(content)
        node = self.top()
        node.charbuffer.append(text)

    def push(self, node):
        self.nodes.append(node)
        return node

    def pop(self):
        return self.nodes.pop()

    def top(self):
        return self.nodes[-1]


class Parser:
    """SAX parser."""

    @classmethod
    def saxparser(cls):
        p = make_parser()
        p.setFeature(feature_external_ges, 0)
        h = Handler()
        p.setContentHandler(h)
        return p, h

    def parse(self, file=None, string=None):
        """
        SAX parse XML text.

        @param file: Parse a python I{file-like} object.
        @type file: I{file-like} object
        @param string: Parse string XML.
        @type string: str
        @return: Parsed XML document.
        @rtype: L{Document}

        """
        if file is None and string is None:
            return
        timer = suds.metrics.Timer()
        timer.start()
        source = file
        if file is None:
            source = InputSource(None)
            source.setByteStream(suds.BytesIO(string))
        sax, handler = self.saxparser()
        sax.parse(source)
        timer.stop()
        if file is None:
            suds.metrics.log.debug("%s\nsax duration: %s", string, timer)
        else:
            suds.metrics.log.debug("sax (%s) duration: %s", file, timer)
        return handler.nodes[0]



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jeff Ortel ( jortel@redhat.com )

"""Classes representing I{built-in} XSD schema objects."""

from suds import *
from suds.xsd import *
from suds.sax.date import *
from suds.xsd.sxbase import XBuiltin

import datetime
import decimal
import sys


class XAny(XBuiltin):
    """Represents an XSD <xsd:any/> node."""

    def __init__(self, schema, name):
        XBuiltin.__init__(self, schema, name)
        self.nillable = False

    def get_child(self, name):
        child = XAny(self.schema, name)
        return child, []

    def any(self):
        return True


class XBoolean(XBuiltin):
    """Represents an XSD boolean built-in type."""

    _xml_to_python = {"1": True, "true": True, "0": False, "false": False}
    _python_to_xml = {True: "true", 1: "true", False: "false", 0: "false"}

    @staticmethod
    def translate(value, topython=True):
        if topython:
            if isinstance(value, str):
                return XBoolean._xml_to_python.get(value)
        else:
            if isinstance(value, (bool, int)):
                return XBoolean._python_to_xml.get(value)
            return value


class XDate(XBuiltin):
    """Represents an XSD <xsd:date/> built-in type."""

    @staticmethod
    def translate(value, topython=True):
        if topython:
            if isinstance(value, str) and value:
                return Date(value).value
        else:
            if isinstance(value, datetime.date):
                return Date(value)
            return value


class XDateTime(XBuiltin):
    """Represents an XSD <xsd:datetime/> built-in type."""

    @staticmethod
    def translate(value, topython=True):
        if topython:
            if isinstance(value, str) and value:
                return DateTime(value).value
        else:
            if isinstance(value, datetime.datetime):
                return DateTime(value)
            return value


class XDecimal(XBuiltin):
    """
    Represents an XSD <xsd:decimal/> built-in type.

    Excerpt from the XSD datatype specification
    (http://www.w3.org/TR/2004/REC-xmlschema-2-20041028):

    > 3.2.3 decimal
    >
    > [Definition:] decimal represents a subset of the real numbers, which can
    > be represented by decimal numerals. The ·value space· of decimal is the
    > set of numbers that can be obtained by multiplying an integer by a
    > non-positive power of ten, i.e., expressible as i × 10^-n where i and n
    > are integers and n >= 0. Precision is not reflected in this value space;
    > the number 2.0 is not distinct from the number 2.00. The ·order-relation·
    > on decimal is the order relation on real numbers, restricted to this
    > subset.
    >
    > 3.2.3.1 Lexical representation
    >
    > decimal has a lexical representation consisting of a finite-length
    > sequence of decimal digits (#x30-#x39) separated by a period as a decimal
    > indicator. An optional leading sign is allowed. If the sign is omitted,
    > "+" is assumed. Leading and trailing zeroes are optional. If the
    > fractional part is zero, the period and following zero(es) can be
    > omitted. For example: -1.23, 12678967.543233, +100000.00, 210.

    """

    # Python versions before 2.7 do not support the decimal.Decimal.canonical()
    # method but they maintain their decimal.Decimal encoded in canonical
    # format internally so we can easily emulate that function by simply
    # returning the same decimal instance.
    if sys.version_info < (2, 7):
        _decimal_canonical = staticmethod(lambda decimal: decimal)
    else:
        _decimal_canonical = decimal.Decimal.canonical

    @staticmethod
    def _decimal_to_xsd_format(value):
        """
        Converts a decimal.Decimal value to its XSD decimal type value.

        Result is a string containing the XSD decimal type's lexical value
        representation. The conversion is done without any precision loss.

        Note that Python's native decimal.Decimal string representation will
        not do here as the lexical representation desired here does not allow
        representing decimal values using float-like `<mantissa>E<exponent>'
        format, e.g. 12E+30 or 0.10006E-12.

        """
        value = XDecimal._decimal_canonical(value)
        negative, digits, exponent = value.as_tuple()

        # The following implementation assumes the following tuple decimal
        # encoding (part of the canonical decimal value encoding):
        #  - digits must contain at least one element
        #  - no leading integral 0 digits except a single one in 0 (if a non-0
        #    decimal value has leading integral 0 digits they must be encoded
        #    in its 'exponent' value and not included explicitly in its
        #    'digits' tuple)
        assert digits
        assert digits[0] != 0 or len(digits) == 1

        result = []
        if negative:
            result.append("-")

        # No fractional digits.
        if exponent >= 0:
            result.extend(str(x) for x in digits)
            result.extend("0" * exponent)
            return "".join(result)

        digit_count = len(digits)

        # Decimal point offset from the given digit start.
        point_offset = digit_count + exponent

        # Trim trailing fractional 0 digits.
        fractional_digit_count = min(digit_count, -exponent)
        while fractional_digit_count and digits[digit_count - 1] == 0:
            digit_count -= 1
            fractional_digit_count -= 1

        # No trailing fractional 0 digits and a decimal point coming not after
        # the given digits, meaning there is no need to add additional trailing
        # integral 0 digits.
        if point_offset <= 0:
            # No integral digits.
            result.append("0")
            if digit_count > 0:
                result.append(".")
                result.append("0" * -point_offset)
                result.extend(str(x) for x in digits[:digit_count])
        else:
            # Have integral and possibly some fractional digits.
            result.extend(str(x) for x in digits[:point_offset])
            if point_offset < digit_count:
                result.append(".")
                result.extend(str(x) for x in digits[point_offset:digit_count])
        return "".join(result)

    @classmethod
    def translate(cls, value, topython=True):
        if topython:
            if isinstance(value, str) and value:
                return decimal.Decimal(value)
        else:
            if isinstance(value, decimal.Decimal):
                return cls._decimal_to_xsd_format(value)
            return value


class XFloat(XBuiltin):
    """Represents an XSD <xsd:float/> built-in type."""

    @staticmethod
    def translate(value, topython=True):
        if topython:
            if isinstance(value, str) and value:
                return float(value)
        else:
            return value


class XInteger(XBuiltin):
    """Represents an XSD <xsd:int/> built-in type."""

    @staticmethod
    def translate(value, topython=True):
        if topython:
            if isinstance(value, str) and value:
                return int(value)
        else:
            return value


class XLong(XBuiltin):
    """Represents an XSD <xsd:long/> built-in type."""

    @staticmethod
    def translate(value, topython=True):
        if topython:
            if isinstance(value, str) and value:
                return int(value)
        else:
            return value


class XString(XBuiltin):
    """Represents an XSD <xsd:string/> node."""
    pass


class XTime(XBuiltin):
    """Represents an XSD <xsd:time/> built-in type."""

    @staticmethod
    def translate(value, topython=True):
        if topython:
            if isinstance(value, str) and value:
                return Time(value).value
        else:
            if isinstance(value, datetime.time):
                return Time(value)
            return value


class Factory:

    tags = {
        # any
        "anyType": XAny,
        # strings
        "string": XString,
        "normalizedString": XString,
        "ID": XString,
        "Name": XString,
        "QName": XString,
        "NCName": XString,
        "anySimpleType": XString,
        "anyURI": XString,
        "NOTATION": XString,
        "token": XString,
        "language": XString,
        "IDREFS": XString,
        "ENTITIES": XString,
        "IDREF": XString,
        "ENTITY": XString,
        "NMTOKEN": XString,
        "NMTOKENS": XString,
        # binary
        "hexBinary": XString,
        "base64Binary": XString,
        # integers
        "int": XInteger,
        "integer": XInteger,
        "unsignedInt": XInteger,
        "positiveInteger": XInteger,
        "negativeInteger": XInteger,
        "nonPositiveInteger": XInteger,
        "nonNegativeInteger": XInteger,
        # longs
        "long": XLong,
        "unsignedLong": XLong,
        # shorts
        "short": XInteger,
        "unsignedShort": XInteger,
        "byte": XInteger,
        "unsignedByte": XInteger,
        # floats
        "float": XFloat,
        "double": XFloat,
        "decimal": XDecimal,
        # dates & times
        "date": XDate,
        "time": XTime,
        "dateTime": XDateTime,
        "duration": XString,
        "gYearMonth": XString,
        "gYear": XString,
        "gMonthDay": XString,
        "gDay": XString,
        "gMonth": XString,
        # boolean
        "boolean": XBoolean,
    }

    @classmethod
    def maptag(cls, tag, fn):
        """
        Map (override) tag => I{class} mapping.

        @param tag: An XSD tag name.
        @type tag: str
        @param fn: A function or class.
        @type fn: fn|class.

        """
        cls.tags[tag] = fn

    @classmethod
    def create(cls, schema, name):
        """
        Create an object based on the root tag name.

        @param schema: A schema object.
        @type schema: L{schema.Schema}
        @param name: The name.
        @type name: str
        @return: The created object.
        @rtype: L{XBuiltin}

        """
        fn = cls.tags.get(name, XBuiltin)
        return fn(schema, name)



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jurko Gospodnetić ( jurko.gospodnetic@pke.hr )

"""
Generic functionality shared in different development utility modules.

"""

import os
import os.path
import sys
if sys.version_info < (3, 0):
    from urllib import quote as url_quote
else:
    from urllib.parse import quote as url_quote


def any_contains_any(strings, candidates):
    """Whether any of the strings contains any of the candidates."""
    for string in strings:
        for c in candidates:
            if c in string:
                return True


class FileJanitor:
    """Janitor class for removing a specific file."""

    def __init__(self, path):
        self.__path = path

    def clean(self):
        try:
            os.remove(self.__path)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            pass


def lowest_version_string_with_prefix(prefix):
    """
    The lowest possible version string with the given prefix.

    'The lowest' according to the usual version string ordering used by
    setuptools, e.g. '2.4.3.dev0' is the lowest possible version in the 2.4.3
    series.

    """
    return "%s.dev0" % (prefix,)


def path_to_URL(path, escape=True):
    """Convert a local file path to a absolute path file protocol URL."""
    # We do not use urllib's builtin pathname2url() function since:
    #  - it has been commented with 'not recommended for general use'
    #  - it does not seem to work the same on Windows and non-Windows platforms
    #    (result starts with /// on Windows but does not on others)
    #  - urllib implementation prior to Python 2.5 used to quote ':' characters
    #    as '|' which would confuse pip on Windows.
    url = os.path.abspath(path)
    for sep in (os.sep, os.altsep):
        if sep and sep != "/":
            url = url.replace(sep, "/")
    if escape:
        # Must not escape ':' or '/' or Python will not recognize those URLs
        # correctly. Detected on Windows 7 SP1 x64 with Python 3.4.0, but doing
        # this always does not hurt since both are valid ASCII characters.
        no_protocol_URL = url_quote(url, safe=":/")
    else:
        no_protocol_URL = url
    return "file:///%s" % (no_protocol_URL,)


def report_error(message):
    print("ERROR: %s" % (message,))


def requirement_spec(package_name, *args):
    """Identifier used when specifying a requirement to pip or setuptools."""
    if not args or args == (None,):
        return package_name
    version_specs = []
    for version_spec in args:
        if isinstance(version_spec, (list, tuple)):
            operator, version = version_spec
        else:
            assert isinstance(version_spec, str)
            operator = "=="
            version = version_spec
        version_specs.append("%s%s" % (operator, version))
    return "%s%s" % (package_name, ",".join(version_specs))


if sys.version_info < (2, 6):
    def _rel_path(path):
        # Older Python versions do not support os.path.relpath(), ergo no
        # pretty path formatting for them.
        return os.path.normpath(path)
else:
    def _rel_path(path):
        try:
            return os.path.relpath(path)
        except ValueError:
            # Current and given path on different file systems, e.g. C: & D:
            # drives on Windows.
            return os.path.normpath(path)


def script_folder(script_path):
    """
    Return the given script's folder or None if it can not be determined.

    Script is identified by its __file__ attribute. If the given __file__
    attribute value contains no path information, it is expected to identify an
    existing file in the current working folder.

    Returned folder may be specified relative to the current working folder
    and, if determined, will never be an empty string.

    Typical use case for calling this function is from a regular stand-alone
    script and not a frozen module or a module imported from the disk, a
    zip-file, an external database or any other such source. Such callers can
    safely assume they have a valid __file__ attribute available.

    """
    # There exist modules whose __file__ attribute does not correspond directly
    # to a disk file, e.g. modules imported from inside zip archives.
    if os.path.isfile(script_path):
        return _rel_path(os.path.dirname(script_path)) or "."


def path_iter(path):
    """Returns an iterator over all the file & folder names in a path."""
    parts = []
    while path:
        path, item = os.path.split(path)
        if item:
            parts.append(item)
    return reversed(parts)



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jurko Gospodnetić ( jurko.gospodnetic@pke.hr )

"""
Suds library HTTP transport related unit tests.

Implemented using the 'pytest' testing framework.

"""

import testutils
if __name__ == "__main__":
    testutils.run_using_pytest(globals())

import suds
import suds.transport
import suds.transport.http

import pytest
from six import u
from six.moves import http_client
from six.moves.urllib.error import HTTPError
from six.moves.urllib.request import ProxyHandler

import base64
import re
import sys
from email.message import Message

# We can not use six.moves modules for this since we want to monkey-patch the
# exact underlying urllib2/urllib.request module in our tests and not just
# their six.moves proxy module.
if sys.version_info < (3,):
    import urllib2 as urllib_request
else:
    import urllib.request as urllib_request


class MustNotBeCalled(Exception):
    """Local exception used in this test module."""
    pass


class MyException(Exception):
    """Local exception used in this test module."""
    pass


class Undefined:
    """Internal tag class indicating undefined function call parameters."""
    pass


class CountedMock(object):
    """
    Base mock object class supporting generic attribute access counting.

    Ignores attributes whose name starts with 'mock_' or '__mock_' or their
    transformed variant '_<className>__mock_'.

    Derived classes must call this class's __init__() or mock_reset() methods
    during their initialization, but both calls are not needed. Before this
    initialization, all counters will be reported as None.

    """

    def __init__(self):
        self.mock_reset()

    def __getattribute__(self, name):
        get = super(CountedMock, self).__getattribute__
        counter_name = "_CountedMock__mock_call_counter"
        has_counter = False
        try:
            counter = get(counter_name)
            has_counter = True
        except AttributeError:
            pass
        if has_counter:
            if name == counter_name:
                return counter
            if not re.match("(_.+__)?mock_", name):
                counter[name] = counter.get(name, 0) + 1
        return get(name)

    def mock_call_count(self, name):
        if self.__mock_call_counter:
            return self.__mock_call_counter.get(name, 0)

    def mock_reset(self):
        self.__mock_call_counter = {}


class MockFP:
    """
    Mock FP 'File object' as stored inside Python's HTTPError exception.

    Must have several 'File object' methods defined on it as Python's HTTPError
    implementation expects them and stores references to them internally, at
    least with Python 2.4.

    """

    def read():
        raise MustNotBeCalled

    def readline():
        raise MustNotBeCalled

    def close(self):
        pass


class MockURLOpenerSaboteur:
    """
    Mock URLOpener raising an exception in its open() method.

    If no open_exception is given in its initializer, simply marks the current
    test as a failure if its open() method is called. Otherwise raises the
    given exception from that call.

    """

    def __init__(self, open_exception=None):
        self.__open_exception = open_exception
        self.args = None

    def open(self, *args, **kwargs):
        self.args = args
        if self.__open_exception:
            raise self.__open_exception
        pytest.fail("urllib urlopener.open() must not be called.")


class SendMethodFixture:
    """
    Instances of this class get returned by the send_method test fixture.

    Each instance is connected to a specific suds.transport.http.HttpTransport
    request sending method and may be used to call that method on a specific
    suds.transport.http.HttpTransport instance.

    """

    def __init__(self, name):
        self.name = name

    def __call__(self, transport, *args, **kwargs):
        assert isinstance(transport, suds.transport.http.HttpTransport)
        return getattr(transport, self.name)(*args, **kwargs)


# Test URL data used by several tests in this test module.
test_URL_data = (
    "sudo://make-me-a-sammich",
    "http://my little URL",
    "https://my little URL",
    "xxx://my little URL",
    "xxx:my little URL",
    "xxx:")


def assert_default_transport(transport):
    """Test utility verifying default constructed transport content."""
    assert isinstance(transport, suds.transport.http.HttpTransport)
    assert transport.urlopener is None


def create_request(url="protocol://default-url", data=u("Rumpelstiltskin")):
    """Test utility constructing a suds.transport.Request instance."""
    return suds.transport.Request(url, data)


@pytest.fixture(params=["open", "send"])
def send_method(request):
    """
    pytest testing framework based test fixture causing tests using that
    fixture to be called for each suds.transport.http.HttpTransport request
    sending method.

    """
    return SendMethodFixture(request.param)


@pytest.mark.parametrize("input", (
    dict(),
    dict(password="Humpty"),
    dict(username="Dumpty"),
    dict(username="Habul Afufa", password="preCious"),
    # Regression test: Extremely long username & password combinations must not
    # cause suds to add additional newlines in the constructed 'Authorization'
    # HTTP header.
    dict(username="An Extremely Long Username that could be usable only to "
        "Extremely Important People whilst on Extremely Important Missions.",
        password="An Extremely Long Password that could be usable only to "
        "Extremely Important People whilst on Extremely Important Missions. "
        "And some extra 'funny' characters to improve security: "
        "!@#$%^&*():|}|{{.\nEven\nSome\nNewLines\n"
        "  and spaces at the start of a new line.   ")))
def test_authenticated_http_add_credentials_to_request(input):
    class MockRequest:
        def __init__(self):
            self.headers = {}

    def assert_Authorization_header(request, username, password):
        if username is None or password is None:
            assert len(request.headers) == 0
        else:
            assert len(request.headers) == 1
            header = request.headers["Authorization"]
            assert header == _encode_basic_credentials(username, password)

    username = input.get("username", None)
    password = input.get("password", None)
    t = suds.transport.http.HttpAuthenticated(**input)
    r = MockRequest()
    t.addcredentials(r)
    assert_Authorization_header(r, username, password)


@pytest.mark.parametrize("input", (
    dict(password="riff raff..!@#"),
    dict(username="macro"),
    dict(username="Hab AfuFa", password="preCious")))
def test_construct_authenticated_http(input):
    expected_username = input.get("username", None)
    expected_password = input.get("password", None)
    transport = suds.transport.http.HttpAuthenticated(**input)
    assert transport.credentials() == (expected_username, expected_password)
    assert_default_transport(transport)


def test_request_headers_are_passed_to_urllib_when_opening_a_request():
    class MockRequest(object):
        url = "http://somewhere.far"
        headers = {'Key': 'value'}

    mock_request = MockRequest()

    transport = suds.transport.http.HttpTransport()
    transport.urlopener = MockURLOpenerSaboteur(MyException)

    pytest.raises(MyException, suds.transport.http.HttpTransport.open, transport, mock_request)

    assert transport.urlopener.args[0].headers == mock_request.headers


def test_construct_http():
    transport = suds.transport.http.HttpTransport()
    assert_default_transport(transport)


def test_sending_using_network_sockets(send_method, monkeypatch):
    """
    Test that telling HttpTransport to send a request actually causes it to
    send the expected data over the network.

    In order to test this we need to make suds attempt to send a HTTP request
    over the network. On the other hand, we want the test to work even on
    computers not connected to a network so we monkey-patch the underlying
    network socket APIs, log all the data suds attempts to send over the
    network and consider the test run successful once suds attempts to read
    back data from the network.

    """

    class Mocker(CountedMock):
        def __init__(self, expected_host, expected_port):
            self.expected_host = expected_host
            self.expected_port = expected_port
            self.host_address = object()
            self.mock_reset()
        def getaddrinfo(self, host, port, *args, **kwargs):
            assert host == self.expected_host
            assert port == self.expected_port
            return [(None, None, None, None, self.host_address)]
        def mock_reset(self):
            super(Mocker, self).mock_reset()
            self.mock_sent_data = suds.byte_str()
            self.mock_socket = None
        def socket(self, *args, **kwargs):
            assert self.mock_socket is None
            self.mock_socket = MockSocket(self)
            return self.mock_socket

    class MockSocket(CountedMock):
        def __init__(self, mocker):
            self.__mocker = mocker
            self.mock_reset()
        def close(self):
            pass
        def connect(self, address):
            assert address is self.__mocker.host_address
        def makefile(self, *args, **kwargs):
            assert self.mock_reader is None
            self.mock_reader = MockSocketReader()
            return self.mock_reader
        def mock_reset(self):
            super(MockSocket, self).mock_reset()
            self.mock_reader = None
        def sendall(self, data):
            self.__mocker.mock_sent_data += data
        def settimeout(self, *args, **kwargs):
            pass
        def setsockopt(self, *args, **kwargs):
            pass

    class MockSocketReader(CountedMock):
        def __init__(self):
            super(MockSocketReader, self).__init__()
        def readline(self, *args, **kwargs):
            raise MyException
        def close(self):
            pass
        def flush(self):
            pass

    # Setup.
    host = "an-easily-recognizable-host-name-214894932"
    port = 9999
    host_port = "%s:%s" % (host, port)
    url_relative = "svc"
    url = "http://%s/%s" % (host_port, url_relative)
    partial_ascii_byte_data = suds.byte_str("Muka-laka-hiki")
    non_ascii_byte_data = u("\u0414\u043C\u0438 \u0442\u0440").encode("utf-8")
    non_ascii_byte_data += partial_ascii_byte_data
    mocker = Mocker(host, port)
    monkeypatch.setattr("socket.getaddrinfo", mocker.getaddrinfo)
    monkeypatch.setattr("socket.socket", mocker.socket)
    request = suds.transport.Request(url, non_ascii_byte_data)
    transport = suds.transport.http.HttpTransport()
    expected_sent_data_start, expected_request_data_send = {
        "open": ("GET", False),
        "send": ("POST", True)}[send_method.name]

    # Execute.
    pytest.raises(MyException, send_method, transport, request)

    # Verify.
    assert mocker.mock_call_count("getaddrinfo") == 1
    assert mocker.mock_call_count("socket") == 1
    assert mocker.mock_socket.mock_call_count("connect") == 1
    assert mocker.mock_socket.mock_call_count("makefile") == 1
    # With older Python versions, e.g. Python 2.4, urllib implementation calls
    # Socket's sendall() method twice - once for sending the HTTP request
    # headers and once for its body.
    assert mocker.mock_socket.mock_call_count("sendall") in (1, 2)
    # Python versions prior to 3.4.2 do not explicitly close their HTTP server
    # connection socket in case of our custom exceptions, e.g. version 3.4.1.
    # closes it only on OSError exceptions.
    assert mocker.mock_socket.mock_call_count("close") in (0, 1)
    # With older Python versions, e.g. Python 2.4, Socket class does not
    # implement the settimeout() method.
    assert mocker.mock_socket.mock_call_count("settimeout") in (0, 1)
    assert mocker.mock_socket.mock_reader.mock_call_count("readline") == 1
    assert mocker.mock_sent_data.__class__ is suds.byte_str_class
    expected_sent_data_start = "%s /%s HTTP/1.1\r\n" % (
        expected_sent_data_start, url_relative)
    expected_sent_data_start = suds.byte_str(expected_sent_data_start)
    assert mocker.mock_sent_data.startswith(expected_sent_data_start)
    assert host_port.encode("utf-8") in mocker.mock_sent_data
    if expected_request_data_send:
        assert mocker.mock_sent_data.endswith(non_ascii_byte_data)
    else:
        assert partial_ascii_byte_data not in mocker.mock_sent_data


class TestSendingToURLWithAMissingProtocolIdentifier:
    """
    Test suds reporting URLs with a missing protocol identifier.

    Python urllib library makes this check under Python 3.x, but not under
    earlier Python versions.

    """

    # We can not set this 'url' fixture data using a class decorator since that
    # Python feature has been introduced in Python 2.6 and we need to keep this
    # code backward compatible with Python 2.4.
    invalid_URL_parametrization = pytest.mark.parametrize("url", (
        "my no-protocol URL",
        ":my no-protocol URL"))

    @pytest.mark.skipif(sys.version_info >= (3,), reason="Python 2 specific")
    @invalid_URL_parametrization
    def test_python2(self, url, send_method):
        transport = suds.transport.http.HttpTransport()
        transport.urlopener = MockURLOpenerSaboteur(MyException)
        request = create_request(url)
        pytest.raises(MyException, send_method, transport, request)

    @pytest.mark.skipif(sys.version_info < (3,), reason="Python 3+ specific")
    @invalid_URL_parametrization
    def test_python3(self, url, send_method, monkeypatch):
        monkeypatch.delitem(locals(), "e", False)
        transport = suds.transport.http.HttpTransport()
        transport.urlopener = MockURLOpenerSaboteur()
        request = create_request(url)
        e = pytest.raises(ValueError, send_method, transport, request)
        try:
            assert "unknown url type" in str(e.value)
        finally:
            del e  # explicitly break circular reference chain in Python 3


class TestURLOpenerUsage:
    """
    Test demonstrating how suds.transport.http.HttpTransport makes use of the
    urllib library to perform the actual network transfers.

    The main contact point with the urllib library are its OpenerDirector
    objects we refer to as 'urlopener'.

    """

    @staticmethod
    def create_HTTPError(url=Undefined, code=Undefined, msg=Undefined,
            hdrs=Undefined, fp=None):
        """
        Test utility method constructing a HTTPError instance. Allows callers
        to construct a HTTPError instance using input data they are interested
        in, with some built-in default values used for any input data they are
        not interested in.

        """
        if url is Undefined:
            url = object()
        if code is Undefined:
            code = object()
        if msg is Undefined:
            msg = object()
        if hdrs is Undefined:
            hdrs = object()
        return HTTPError(url=url, code=code, msg=msg, hdrs=hdrs, fp=fp)

    @pytest.mark.parametrize("status_code", (
        http_client.ACCEPTED,
        http_client.NO_CONTENT,
        http_client.RESET_CONTENT,
        http_client.MOVED_PERMANENTLY,
        http_client.BAD_REQUEST,
        http_client.PAYMENT_REQUIRED,
        http_client.FORBIDDEN,
        http_client.NOT_FOUND,
        http_client.INTERNAL_SERVER_ERROR,
        http_client.NOT_IMPLEMENTED,
        http_client.HTTP_VERSION_NOT_SUPPORTED))
    def test_open_propagating_HTTPError_exceptions(self, status_code,
            monkeypatch):
        """
        HttpTransport open() operation should transform HTTPError urlopener
        exceptions to suds.transport.TransportError exceptions.

        """
        # Setup.
        monkeypatch.delattr(locals(), "e", False)
        fp = MockFP()
        e_original = self.create_HTTPError(code=status_code, fp=fp)
        t = suds.transport.http.HttpTransport()
        t.urlopener = MockURLOpenerSaboteur(open_exception=e_original)
        request = create_request()

        # Execute.
        e = pytest.raises(suds.transport.TransportError, t.open, request).value
        try:
            # Verify.
            assert e.args == (str(e_original),)
            assert e.httpcode is status_code
            assert e.fp is fp
        finally:
            del e  # explicitly break circular reference chain in Python 3

    @pytest.mark.xfail(reason="original suds library bug")
    @pytest.mark.parametrize("status_code", (
        http_client.ACCEPTED,
        http_client.NO_CONTENT))
    def test_operation_invoke_with_urlopen_accept_no_content__data(self,
            status_code):
        """
        suds.client.Client web service operation invocation expecting output
        data, and for which a corresponding urlopen call raises a HTTPError
        with status code ACCEPTED or NO_CONTENT, should report this as a
        TransportError.

        """
        e = self.create_HTTPError(code=status_code)
        transport = suds.transport.http.HttpTransport()
        transport.urlopener = MockURLOpenerSaboteur(open_exception=e)
        wsdl = testutils.wsdl('<xsd:element name="o" type="xsd:string"/>',
            output="o", operation_name="f")
        client = testutils.client_from_wsdl(wsdl, transport=transport)
        pytest.raises(suds.transport.TransportError, client.service.f)

    @pytest.mark.xfail(reason="original suds library bug")
    @pytest.mark.parametrize("status_code", (
        http_client.ACCEPTED,
        http_client.NO_CONTENT))
    def test_operation_invoke_with_urlopen_accept_no_content__no_data(self,
            status_code):
        """
        suds.client.Client web service operation invocation expecting no output
        data, and for which a corresponding urlopen call raises a HTTPError
        with status code ACCEPTED or NO_CONTENT, should treat this as a
        successful invocation.

        """
        # We are not yet sure that the behaviour checked for in this test is
        # actually desired. The test is only an 'educated guess' prepared to
        # demonstrate a related problem in the original suds library
        # implementation. The original implementation is definitely buggy as
        # its web service operation invocation raises an AttributeError
        # exception by attempting to access a non-existing 'None.message'
        # attribute internally.
        e = self.create_HTTPError(code=status_code)
        transport = suds.transport.http.HttpTransport()
        transport.urlopener = MockURLOpenerSaboteur(open_exception=e)
        wsdl = testutils.wsdl('<xsd:element name="o" type="xsd:string"/>',
            output="o", operation_name="f")
        client = testutils.client_from_wsdl(wsdl, transport=transport)
        assert client.service.f() is None

    def test_propagating_non_HTTPError_exceptions(self, send_method):
        """
        HttpTransport data sending operations need to propagate non-HTTPError
        exceptions raised by the underlying urlopen call.

        """
        e = MyException()
        t = suds.transport.http.HttpTransport()
        t.urlopener = MockURLOpenerSaboteur(open_exception=e)
        assert pytest.raises(e.__class__, t.open, create_request()).value is e

    @pytest.mark.parametrize("status_code", (
        http_client.RESET_CONTENT,
        http_client.MOVED_PERMANENTLY,
        http_client.BAD_REQUEST,
        http_client.PAYMENT_REQUIRED,
        http_client.FORBIDDEN,
        http_client.NOT_FOUND,
        http_client.INTERNAL_SERVER_ERROR,
        http_client.NOT_IMPLEMENTED,
        http_client.HTTP_VERSION_NOT_SUPPORTED))
    def test_send_transforming_HTTPError_exceptions(self, status_code,
            monkeypatch):
        """
        HttpTransport send() operation should transform HTTPError urlopener
        exceptions with status codes other than ACCEPTED or NO_CONTENT to
        suds.transport.TransportError exceptions.

        """
        # Setup.
        monkeypatch.delattr(locals(), "e", False)
        msg = object()
        fp = MockFP()
        e_original = self.create_HTTPError(msg=msg, code=status_code, fp=fp)
        t = suds.transport.http.HttpTransport()
        t.urlopener = MockURLOpenerSaboteur(open_exception=e_original)
        request = create_request()

        # Execute.
        e = pytest.raises(suds.transport.TransportError, t.send, request).value
        try:
            # Verify.
            assert len(e.args) == 1
            assert e.args[0] is e_original.msg
            assert e.httpcode is status_code
            assert e.fp is fp
        finally:
            del e  # explicitly break circular reference chain in Python 3

    @pytest.mark.parametrize("status_code", (
        http_client.ACCEPTED,
        http_client.NO_CONTENT))
    def test_send_transforming_HTTPError_exceptions__accepted_no_content(self,
            status_code):
        """
        HttpTransport send() operation should return None when their underlying
        urlopen operation raises a HTTPError exception with status code
        ACCEPTED or NO_CONTENT.

        """
        e_original = self.create_HTTPError(code=status_code)
        t = suds.transport.http.HttpTransport()
        t.urlopener = MockURLOpenerSaboteur(open_exception=e_original)
        assert t.send(create_request()) is None

    def test_specify_timeout(self):
        """
        HttpTransport send() operation should pass a Request timeout parameter
        to urllib
        """
        t = suds.transport.http.HttpTransport()
        request = create_request()
        request.timeout = 10

        # Python 2 compatible object
        class CompatibleHeaders(dict):
            dict = {}

        class MockResponse:
            def info(self):
                message = Message()
                # Python 2 compatible response
                message.getheaders = lambda k: {}
                return message

            @property
            def headers(self):
                return CompatibleHeaders()

            def read(self):
                return ''

        class MockURLOpener:
            def open(self, urllib_request, timeout=None):
                assert timeout == request.timeout
                return MockResponse()

        t.urlopener = MockURLOpener()
        t.send(request)


    @pytest.mark.parametrize("url", test_URL_data)
    def test_urlopener_default(self, url, send_method, monkeypatch):
        """
        HttpTransport builds a new urlopener if not given an external one.

        """
        def my_build_urlopener(*handlers):
            assert len(handlers) == 1
            assert handlers[0].__class__ is ProxyHandler
            raise MyException
        monkeypatch.setattr(urllib_request, "build_opener", my_build_urlopener)
        transport = suds.transport.http.HttpTransport()
        request = create_request(url=url)
        pytest.raises(MyException, send_method, transport, request)

    @pytest.mark.parametrize("url", test_URL_data)
    def test_urlopener_indirection(self, url, send_method, monkeypatch):
        """
        HttpTransport may be configured with an external urlopener.

        In that case, when opening or sending a HTTP request, a new urlopener
        is not built and the given urlopener is used as-is, without adding any
        extra handlers to it.

        """
        class MockURLOpener:
            def open(self, request, timeout=None):
                assert request.__class__ is urllib_request.Request
                assert request.get_full_url() == url
                raise MyException
        transport = suds.transport.http.HttpTransport()
        transport.urlopener = MockURLOpener()
        def my_build_urlopener(*args, **kwargs):
            pytest.fail("urllib build_opener() called when not expected.")
        monkeypatch.setattr(urllib_request, "build_opener", my_build_urlopener)
        request = create_request(url=url)
        pytest.raises(MyException, send_method, transport, request)


def _encode_basic_credentials(username, password):
    """
    Encode user credentials as used in basic HTTP authentication.

    This is the value expected to be added to the 'Authorization' HTTP header.

    """
    data = suds.byte_str("%s:%s" % (username, password))
    return "Basic %s" % base64.b64encode(data).decode("utf-8")



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jurko Gospodnetić ( jurko.gospodnetic@pke.hr )

"""
Indirect parametrization pytest plugin.

Allows tests parametrization data to be generated using a function instead of
having to be explicitly specified as is the case when using the builtin
pytest.mark.parametrize functionality.

Interface is similar to the builtin pytest.mark.parametrize implementation,
except that it takes an indirect parametrization function as an additional
initial positional argument. All the other arguments parameters are forwarded
onto this indirect parametrization function which then calculates and returns
the actual parametrization data. The return values are a standard Python
positional argument list and keyword argument dictionary to be used for the
underlying metafunc.parametrize() call.

May be used as either:
  1. pytest.indirect_parametrizer() or
  2. pytest.mark.indirect_parametrizer().

The two usages are equivalent except that the first one may be used with any
indirect parametrization function while the second one can not be used with
indirect parametrization functions taking no input parameters (with pytest
versions prior to 2.5.2 it also can not be used with functions taking only
keyword arguments). This is a technical restriction based on an underlying
pytest.mark implementation detail. See the indirect_parametrize() function
doc-string for more detailed information.

Usage example making the following test_example() test function calls after
calculating them from a much shorter definition:
 * test_example("Fritula", 1)
 * test_example("Fritula", 2)
 * test_example("Fritula", 3)
 * test_example("Fritula", 4)
 * test_example("Fritula", 5)
 * test_example("Fritula", 6)
 * test_example("Fritula", 7)
 * test_example("Fritula", 8)
 * test_example("Fritula", 9)
 * test_example("Madagascar", 10)
 * test_example("Madagascar", 20)
 * test_example("Madagascar", 30)
 * test_example("Rumpelstiltskin", 20)
 * test_example("Rumpelstiltskin", 40)

def custom_parametrization(param_names, param_value_defs):
    param_values = []
    for uno, due_values in param_value_defs:
        for due in due_values:
            param_values.append((uno, due))
    return (param_names, param_values), {}

@pytest.indirect_parametrize(custom_parametrization, ("uno", "due"), (
    ("Fritula", (1, 2, 3, 4, 5, 6, 7, 8, 9)),
    ("Madagascar", (10, 30, 50)),
    ("Rumpelstiltskin", (20, 40))))
def test_example(uno, due):
    assert False

"""

import pytest


def indirect_parametrize(func, *args, **kwargs):
    """
    Decorator registering a custom parametrization function for a pytest test.

    This pytest.mark.indirect_parametrize() replacement allows the use of
    indirect parametrization functions taking no input parameters or, with
    pytest versions prior to 2.5.2, functions taking only keyword arguments.

    If a pytest.mark.indirect_parametrize() call is made with such an indirect
    parametrization function, it decorates the given function instead of
    storing and using it to decorate the intended function later on.

    """
    # In pytest versions prior to 2.5.2 pytest.mark.indirect_parametrize()
    # special handling occurs even when passing it additional keyword arguments
    # so we have to make sure we are passing it at least one additional
    # positional argument.
    def wrapper(func, *args, **kwargs):
        return func(*args, **kwargs)
    return pytest.mark.indirect_parametrize(wrapper, func, *args, **kwargs)


def pytest_configure(config):
    """Describe the new pytest marker in the --markers output."""
    config.addinivalue_line("markers",
        "indirect_parametrize(function, argnames, argvalues): similar to the "
        "builtin pytest.mark.parametrize implementation, except that it takes "
        "an indirect parametrization function as an additional initial "
        "positional argument. All the other parameters are forwarded to the "
        "indirect parametrization function which then returns the actual "
        "metafunc.parametrize() parameters (standard Python positional "
        "argument list and keyword argument dictionary) based on the received "
        "input data. For more detailed information see the "
        "indirect_parametrize pytest plugin implementation module.")
    """pytest hook publishing references in the toplevel pytest namespace."""
    pytest.indirect_parametrize = indirect_parametrize


def pytest_generate_tests(metafunc):
    """pytest hook called for all detected test functions."""
    mark = metafunc.definition.get_closest_marker('indirect_parametrize')
    if not mark:
        return
    args, kwargs = mark.args[0](*mark.args[1:], **mark.kwargs)
    metafunc.parametrize(*args, **kwargs)




# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jurko Gospodnetić ( jurko.gospodnetic@pke.hr )

"""
Utility for converting version strings into a directly comparable tuple.

Shamelessly stolen from setuptools 3.7 (development version) where this code is
located in its pkg_resources.py module. The code has not been modified in any
way, including comments and coding style, so that it would be as easy as
possible to upgrade it to a newer version if that becomes needed.

Having this functionality in our project allows our tool scripts to run even
without locally installed setuptools installed, e.g. the script for setting up
the base development environments.

"""

import re

component_re = re.compile(r'(\d+ | [a-z]+ | \.| -)', re.VERBOSE)
replace = {'pre':'c', 'preview':'c','-':'final-','rc':'c','dev':'@'}.get

def _parse_version_parts(s):
    for part in component_re.split(s):
        part = replace(part, part)
        if not part or part=='.':
            continue
        if part[:1] in '0123456789':
            # pad for numeric comparison
            yield part.zfill(8)
        else:
            yield '*'+part

    # ensure that alpha/beta/candidate are before final
    yield '*final'

def parse_version(s):
    """Convert a version string to a chronologically-sortable key

    This is a rough cross between distutils' StrictVersion and LooseVersion;
    if you give it versions that would work with StrictVersion, then it behaves
    the same; otherwise it acts like a slightly-smarter LooseVersion. It is
    *possible* to create pathological version coding schemes that will fool
    this parser, but they should be very rare in practice.

    The returned value will be a tuple of strings.  Numeric portions of the
    version are padded to 8 digits so they will compare numerically, but
    without relying on how numbers compare relative to strings.  Dots are
    dropped, but dashes are retained.  Trailing zeros between alpha segments
    or dashes are suppressed, so that e.g. "2.4.0" is considered the same as
    "2.4". Alphanumeric parts are lower-cased.

    The algorithm assumes that strings like "-" and any alpha string that
    alphabetically follows "final"  represents a "patch level".  So, "2.4-1"
    is assumed to be a branch or patch of "2.4", and therefore "2.4.1" is
    considered newer than "2.4-1", which in turn is newer than "2.4".

    Strings like "a", "b", "c", "alpha", "beta", "candidate" and so on (that
    come before "final" alphabetically) are assumed to be pre-release versions,
    so that the version "2.4" is considered newer than "2.4a1".

    Finally, to handle miscellaneous cases, the strings "pre", "preview", and
    "rc" are treated as if they were "c", i.e. as though they were release
    candidates, and therefore are not as new as a version string that does not
    contain them, and "dev" is replaced with an '@' so that it sorts lower than
    than any other pre-release tag.
    """
    parts = []
    for part in _parse_version_parts(s.lower()):
        if part.startswith('*'):
            # remove '-' before a prerelease tag
            if part<'*final':
                while parts and parts[-1]=='*final-': parts.pop()
            # remove trailing zeros from each series of numeric parts
            while parts and parts[-1]=='00000000':
                parts.pop()
        parts.append(part)
    return tuple(parts)



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jurko Gospodnetić ( jurko.gospodnetic@pke.hr )

"""
CompareSAX test utility class implementation.

Extracted into a separate module, named so pytest would apply its assertion
rewriting to it.

'pytest' assertion rewriting allows CompareSAX to use Python assertion based
XML mismatch reporting and have it work even when run with Python assertions
disabled.

"""

import suds.sax.document
import suds.sax.element
import suds.sax.parser

from six import text_type, u

import sys


class CompareSAX:
    """
    Support for comparing SAX XML structures.

    Not intended to be perfect, but only good enough XML comparison to be used
    internally inside the project's test suite.

    Raw XML data is parsed using a suds SAX parser and the resulting DOM
    structure compared. This means that any XML data differences lost during
    SAX parsing can not be detected. Some examples:
      - all textual content for a single node is concatenated together, so the
        following two XML data segments are considered equivalent:
          1. '<a>xx<b/>yy</a>'
          2. '<a>xxyy<b/></a>'
      - all leading and trailing whitespace is trimmed from textual content for
        nodes having at least one child element, so most XML indentation
        information is lost and the following two XML data segments are
        considered equivalent:
          1. '<a>   <b/>     </a>'
          2. '<a><b/></a>'

    Suds may generate different SOAP request XML data for the same input based
    on the order in which it reads some of its internal data held inside
    unordered containers, e.g. set or dictionary. To compensate for this we
    consider both namespace prefix & namespace declaration (either extra
    declarations or declaration placement) differences in XML documents as
    irrelevant. We do however compare each XML node's namespace, i.e. that
    their namespace names match even if their namespace prefixes do not or if
    those namespaces have been declared on different XML elements.

    """

    def __init__(self):
        self.__context = []

    @staticmethod
    def assertions_enabled():
        """
        Returns whether Python assertions have been enabled in this module.

        CompareSAX class uses Python assertions to report failed comparison
        results so this information in order to know whether related tests
        should be disabled.

        """
        try:
            assert False
        except AssertionError:
            return True
        return False

    @classmethod
    def document2document(cls, lhs, rhs):
        """Compares two SAX XML documents."""
        self = cls()
        try:
            self.__document2document(lhs, rhs, context="document2document")
        except Exception:
            self.__report_context()
            raise

    @classmethod
    def document2element(cls, document, element):
        """
        Compares a SAX XML document structure to a SAX XML element.

        The given document & element are considered successfully matched if the
        document consists of a single XML element matching the given one.

        """
        self = cls()
        self.__push_context("document2element")
        try:
            assert document.__class__ is suds.sax.document.Document
            assert element.__class__ is suds.sax.element.Element
            assert len(document.getChildren()) == 1
            self.__element2element(document.getChildren()[0], element)
        except Exception:
            self.__report_context()
            raise

    @classmethod
    def element2element(cls, lhs, rhs):
        """Compares two SAX XML elements."""
        self = cls()
        self.__push_context("element2element")
        try:
            self.__element2element(lhs, rhs)
        except Exception:
            self.__report_context()
            raise

    @classmethod
    def data2data(cls, lhs, rhs):
        """Compares two SAX XML documents given as strings or bytes objects."""
        self = cls()
        try:
            lhs_doc = self.__parse_data(lhs)
            rhs_doc = self.__parse_data(rhs)
            self.__document2document(lhs_doc, rhs_doc, context="data2data")
        except Exception:
            self.__report_context()
            raise

    @classmethod
    def document2data(cls, lhs, rhs):
        """
        Compares two SAX XML documents, second one given as a string or a bytes
        object.

        """
        self = cls()
        try:
            rhs_doc = self.__parse_data(rhs)
            self.__document2document(lhs, rhs_doc, context="document2data")
        except Exception:
            self.__report_context()
            raise

    def __compare_child_elements(self, lhs, rhs):
        """Compares the given entities' child elements."""
        assert len(lhs.getChildren()) == len(rhs.getChildren())
        count = len(lhs.getChildren())
        for i, (l, r) in enumerate(zip(lhs.getChildren(), rhs.getChildren())):
            self.__element2element(l, r, context_info=(i, count))

    def __compare_element_namespace(self, lhs, rhs):
        """
        Compares the given elements' namespaces.

        Empty string & None XML element namespaces are considered the same to
        compensate for the suds SAX document model representing the following
        'default namespace' scenarios differently:
          <a/>
          <a xmlns=""/>
          <ns:a xmlns:ns=""/>

        """
        #TODO: Make suds SAX element model consistently represent empty/missing
        # namespaces and then update both this method and its docstring.
        self.__push_context("namespace")
        lhs_namespace = lhs.namespace()[1]
        rhs_namespace = rhs.namespace()[1]
        if not lhs_namespace:
            lhs_namespace = None
        if not rhs_namespace:
            rhs_namespace = None
        assert lhs_namespace == rhs_namespace
        self.__pop_context()

    def __compare_element_text(self, lhs, rhs):
        """
        Compares the given elements' textual content.

        Empty string & None XML element texts are considered the same to
        compensate for different XML object tree construction methods
        representing 'no text' elements differently, e.g. depending on whether
        a particular SAX parsed XML element had any whitespace characters in
        its textual data or whether the element got constructed in code to
        represent a SOAP request.

        """
        #TODO: Make suds SAX element model consistently represent empty/missing
        # text content and then update both this method and its docstring.
        self.__push_context("text")
        lhs_text = lhs.text
        rhs_text = rhs.text
        if not lhs_text:
            lhs_text = None
        if not rhs_text:
            rhs_text = None
        assert lhs_text == rhs_text
        self.__pop_context()

    def __document2document(self, lhs, rhs, context):
        """
        Internal document2document comparison worker.

        See document2document() docstring for more detailed information.

        """
        self.__push_context(context)
        assert lhs.__class__ is suds.sax.document.Document
        assert rhs.__class__ is suds.sax.document.Document
        self.__compare_child_elements(lhs, rhs)
        self.__pop_context()

    @staticmethod
    def __element_name(element):
        """Returns a given SAX element's name as unicode or '???' on error."""
        try:
            return text_type(element.name)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            return u("???")

    def __element2element(self, lhs, rhs, context_info=(0, 1)):
        """
        Internal element2element comparison worker.

        See element2element() docstring for more detailed information.

        The context information is an (n, count) 2-element collection
        indicating which element2element() call ('n') this is in a sequence of
        such calls ('count'). The exact context string is constructed based on
        the given elements' names and context information.

        """
        context = self.__element2element_context(lhs, rhs, context_info)
        self.__push_context(context)
        assert lhs.__class__ is suds.sax.element.Element
        assert rhs.__class__ is suds.sax.element.Element
        assert lhs.name == rhs.name
        self.__compare_element_namespace(lhs, rhs)
        self.__compare_element_text(lhs, rhs)
        self.__compare_child_elements(lhs, rhs)
        self.__pop_context()

    @classmethod
    def __element2element_context(cls, lhs, rhs, context_info):
        """
        Return a context string for a given element2element call.

        See the __element2element() docstring for more detailed information.

        """
        n, count = context_info
        assert 0 <= n < count, "broken CompareSAX implementation"
        context_lhs_name = cls.__element_name(lhs)
        context_rhs_name = cls.__element_name(rhs)
        if context_lhs_name == context_rhs_name:
            context_name = context_lhs_name
        else:
            context_name = "%s/%s" % (context_lhs_name, context_rhs_name)
        if count == 1:
            return "<%s>" % (context_name,)
        return "<%s(%d/%d)>" % (context_name, n + 1, count)

    @staticmethod
    def __parse_data(data):
        """
        Construct a SAX XML document based on its data given as a string or a
        bytes object.

        """
        if isinstance(data, text_type):
            data = data.encode("utf-8")
        return suds.sax.parser.Parser().parse(string=data)

    def __pop_context(self):
        self.__context.pop()

    def __push_context(self, context):
        self.__context.append(context)

    def __report_context(self):
        if self.__context:
            sys.stderr.write("Failed SAX XML comparison context:\n")
            sys.stderr.write("  %s\n" % (".".join(self.__context)))



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify it under
# the terms of the (LGPL) GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library Lesser General Public License
# for more details at ( http://www.gnu.org/licenses/lgpl.html ).
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# written by: Jurko Gospodnetić ( jurko.gospodnetic@pke.hr )

"""
Suds library's XSD Element node unit tests.

Implemented using the 'pytest' testing framework.

"""

if __name__ == "__main__":
    import testutils
    testutils.run_using_pytest(globals())


import suds
import suds.options
import suds.sax.parser
import suds.store
import suds.xsd.schema

import pytest
from six import b


# shared test input data
form_test_values = (None, "qualified", "unqualified", "invalid", "")


class TestElementForm:
    """Test whether specific XSD elements are considered qualified."""

    @pytest.mark.parametrize("form_default, form", [(x, y)
        for x in form_test_values
        for y in form_test_values])
    def test_local_element(self, form_default, form):
        parent_name = "Parent"
        element_name = "Elemento"
        namespace = "tns"
        schema_xml = """\
<?xml version='1.0' encoding='UTF-8'?>
<schema xmlns="http://www.w3.org/2001/XMLSchema"
        targetNamespace="%(namespace)s"%(form_default)s>
    <element name="%(parent_name)s">
        <complexType>
            <sequence>
                <element name="%(element_name)s"%(form)s/>
            </sequence>
        </complexType>
    </element>
</schema>""" % {
            "element_name": element_name,
            "form": _attribute_xml("form", form),
            "form_default": _attribute_xml("elementFormDefault", form_default),
            "namespace": namespace,
            "parent_name": parent_name}
        expected = form == "qualified" or (
            (form is None) and (form_default == "qualified"))
        schema = _parse_schema_xml(b(schema_xml))
        parent_element = schema.elements[parent_name, namespace]
        element = parent_element.get_child(element_name)[0]
        assert bool(element.form_qualified) == expected

    @pytest.mark.parametrize("form_default, form, form_referenced", [(x, y, z)
        for x in form_test_values
        for y in form_test_values
        for z in form_test_values])
    def test_reference_to_internal(self, form_default, form, form_referenced):
        """Reference element to an element in the same schema."""
        referenced_name = "Referenced"
        referencing_parent_name = "Referencing"
        namespace = "tns"
        schema_xml = """\
<?xml version='1.0' encoding='UTF-8'?>
<schema xmlns="http://www.w3.org/2001/XMLSchema"
        xmlns:tns="%(namespace)s"
        targetNamespace="%(namespace)s"%(form_default)s>
    <element name="%(referenced_name)s"%(form_referenced)s/>
    <element name="%(referencing_parent_name)s">
        <complexType>
            <sequence>
                <element ref="tns:%(referenced_name)s"%(form_referencing)s/>
            </sequence>
        </complexType>
    </element>
</schema>""" % {
            "form_default": _attribute_xml("elementFormDefault", form_default),
            "form_referenced": _attribute_xml("form", form_referenced),
            "form_referencing": _attribute_xml("form", form),
            "namespace": namespace,
            "referenced_name": referenced_name,
            "referencing_parent_name": referencing_parent_name}
        schema = _parse_schema_xml(b(schema_xml))
        parent_element = schema.elements[referencing_parent_name, namespace]
        referencing_element = parent_element.get_child(referenced_name)[0]
        assert referencing_element.form_qualified

    def test_reference_to_external(self):
        """Reference element to an element in an external schema."""
        schema_xml_here = """\
<?xml version='1.0' encoding='UTF-8'?>
<schema xmlns="http://www.w3.org/2001/XMLSchema" xmlns:there="ns-there">
    <import namespace="ns-there" schemaLocation="suds://there.xsd"/>
    <element name="Referencing">
        <complexType>
            <sequence>
                <element ref="there:Referenced"/>
            </sequence>
        </complexType>
    </element>
</schema>"""
        schema_xml_there = """\
<?xml version='1.0' encoding='UTF-8'?>
<schema xmlns="http://www.w3.org/2001/XMLSchema" targetNamespace="ns-there">
    <element name="Referenced"/>
</schema>"""
        store = suds.store.DocumentStore({"there.xsd": b(schema_xml_there)})
        schema = _parse_schema_xml(b(schema_xml_here), store)
        referenced_element = schema.elements["Referenced", "ns-there"]
        referencing_parent = schema.elements["Referencing", None]
        referencing_element = referencing_parent.get_child("Referenced")[0]
        assert referencing_element.form_qualified

    @pytest.mark.parametrize("form_default, form", [(x, y)
        for x in form_test_values
        for y in form_test_values])
    def test_top_level_element(self, form_default, form):
        element_name = "Elemento"
        namespace = "tns"
        schema_xml = """\
<?xml version='1.0' encoding='UTF-8'?>
<schema xmlns="http://www.w3.org/2001/XMLSchema"
        targetNamespace="%(namespace)s"%(form_default)s>
    <element name="%(element_name)s"%(form)s/>
</schema>""" % {
            "element_name": element_name,
            "form": _attribute_xml("form", form),
            "form_default": _attribute_xml("elementFormDefault", form_default),
            "namespace": namespace}
        schema = _parse_schema_xml(b(schema_xml))
        element = schema.elements[element_name, namespace]
        assert element.form_qualified


def test_reference():
    """Reference to an element in a different schema."""
    schema_xml_here = """\
<?xml version='1.0' encoding='UTF-8'?>
<schema xmlns="http://www.w3.org/2001/XMLSchema" xmlns:there="ns-there">
    <import namespace="ns-there" schemaLocation="suds://there.xsd"/>
    <element name="Referencing">
        <complexType>
            <sequence>
                <element ref="there:Referenced"/>
            </sequence>
        </complexType>
    </element>
</schema>"""
    schema_xml_there = """\
<?xml version='1.0' encoding='UTF-8'?>
<schema xmlns="http://www.w3.org/2001/XMLSchema" targetNamespace="ns-there">
    <element name="Referenced"/>
</schema>"""
    store = suds.store.DocumentStore({"there.xsd": b(schema_xml_there)})
    schema = _parse_schema_xml(b(schema_xml_here), store)
    referenced_element = schema.elements["Referenced", "ns-there"]
    referencing_parent = schema.elements["Referencing", None]
    referencing_element = referencing_parent.get_child("Referenced")[0]
    assert referenced_element.ref is None
    assert referencing_element.ref == ("Referenced", "ns-there")


###############################################################################
#
# Test utilities.
#
###############################################################################

def _attribute_xml(name, value):
    if value is not None:
        return ' %s="%s"' % (name, value)
    return ""


def _parse_schema_xml(xml, documentStore=None):
    """Test utility constructing an XSD schema model from the given XML."""
    parser = suds.sax.parser.Parser()
    document = parser.parse(string=xml)
    root = document.root()
    url = "somewhere://over.the/rainbow"
    options_kwargs = {}
    if documentStore:
        options_kwargs.update(documentStore=documentStore)
    options = suds.options.Options(**options_kwargs)
    return suds.xsd.schema.Schema(root, url, options)
